/* Copyright (c) 2000-2006 ActiveState Software Inc.
   See the file LICENSE.txt for licensing information. */

// A project extension to manage source code control operations in Komodo.

if (typeof(ko)=='undefined') {
    var ko = {};
}
if (typeof(ko.projects)=='undefined') {
    ko.projects = {};
}

(function() {

var log = ko.logging.getLogger('peSCC');
var _prefs = Components.classes["@activestate.com/koPrefService;1"].
                    getService(Components.interfaces.koIPrefService).prefs;
var _obSvc = Components.classes["@mozilla.org/observer-service;1"].
             getService(Components.interfaces.nsIObserverService);

//---- Registry of active SCC services
// 'sccModules' is a dictionary of active SCC services conforming to the
// virtual (so virtual we have defined a common base class) koISCC interface.
// Generally there is one such service for each SCC system (e.g. CVS) that
// Komodo supports.
//
// "Active" means that the SCC system is (1) functional (cannot support CVS
// if CVS is not installed on the user's machine) and (2) enabled (user
// preference). Monitoring (2) can be done via Komodo pref observer system.
// Monitoring (1) is done by observing the 'subsystem_changed'
// notification.
//
var sccModules = {};

function _activateService(type) {
    sccModules[type] = null;
    if (!_prefs.getBooleanPref(type+"Enabled")) return;
    var sccModule = Components.classes["@activestate.com/koSCC?type="+type+";1"].
                  getService(Components.interfaces.koISCC);
    if (sccModule.isFunctional) {
        sccModules[type] = sccModule;
    }
}

function _initializeServices() {
    // use the category manager to find all available scc modules and
    // then try to activate them
    var catman = Components.classes["@mozilla.org/categorymanager;1"]
         .getService(Components.interfaces.nsICategoryManager);
    var category = "category-komodo-scc";
    var scc = catman.enumerateCategory(category);
    while (scc.hasMoreElements()) {
        var entry = scc.getNext().QueryInterface(Components.interfaces.nsISupportsCString).data;
        _activateService(entry);
    }
}

function _SCCObserver() {
    _obSvc.addObserver(this, 'subsystem_changed',false);
    _obSvc.addObserver(this, 'scc_status_message',false);
}
_SCCObserver.prototype.observe = function(subject, topic, data)
{
    // The "subsystem_changed" notification indicates that a Komodo
    // subsystem has changed. "data" indentifies the subsystem. If data is
    // 'p4' or 'cvs' then we might need to update the list of active SCC
    // services.
    switch (topic) {
    case 'scc_status_message':
        ko.scc.logger.append(data);
        break;
    case 'subsystem_changed':
        _activateService(data);
        break;
    }
}

function _SCCPrefObserver() {
    _prefs.addObserver(this);
};
_SCCPrefObserver.prototype.observe = function(prefSet, prefSetID, prefName)
{
    //dump("peSCC: observed pref notification: prefName='"+prefName+"'\n");
    // prefname is typeEnabled
    if (prefName.indexOf("Enabled") > 0) {
        var type = prefName.substr(0, prefName.length - 7);
        if (sccModules.indexOf(type) != -1)
            _activateService(type);
    }
};

// Setup dynamic registration of active SCC services.
_initializeServices()
var _sccObserver = new _SCCObserver();
var _sccPrefObserver = new _SCCPrefObserver();

//---- SCC project extension implementation.

function peSCC() {
    this.name = 'SCC';
    //log.setLevel(ko.logging.LOG_DEBUG);
    this.sysutils = Components.classes["@activestate.com/koSysUtils;1"].
              getService(Components.interfaces.koISysUtils);
    this._lastErrorSvc = Components.classes["@activestate.com/koLastErrorService;1"]
               .getService(Components.interfaces.koILastErrorService);
}

// The following two lines ensure proper inheritance (see Flanagan, p. 144).
peSCC.prototype.constructor = peSCC;

peSCC.prototype.init = function() {
    // register our command handlers
    this.registerCommands();

    // this adds columns to the project/toolbox tree's
    var em = ko.projects.extensionManager;
    em.setDatapoint('Status','sccStatus');
    em.setDatapoint('Rev','sccLocalRevision');
    em.setDatapoint('Depot Rev','sccDepotRevision');
    em.setDatapoint('Action','sccAction');

}

peSCC.prototype.registerCommands = function() {
    var em = ko.projects.extensionManager;
    em.registerCommand('cmd_SCCedit', this);
    em.registerCommand('cmd_SCCadd', this);
    em.registerCommand('cmd_SCCremove', this);
    em.registerCommand('cmd_SCCupdate', this);
    em.registerCommand('cmd_SCCcommit', this);
    em.registerCommand('cmd_SCCdiff', this);
    em.registerCommand('cmd_SCChistory', this);
    em.registerCommand('cmd_SCCrevert', this);

    em.registerCommand('cmd_SCCadd_folder', this);
    em.registerCommand('cmd_SCCupdate_folder', this);
    em.registerCommand('cmd_SCCcommit_folder', this);
    em.registerCommand('cmd_SCCdiff_folder', this);
    em.registerCommand('cmd_SCCrevert_folder', this);
}

peSCC.prototype.registerEventHandlers = function() {
}

peSCC.prototype.registerMenus = function() {
    var menupopup = document.getElementById('menu_SCCmenu_context');
    if (!menupopup) {
        log.error("Couldn't get 'menu_SCCmenu'");
        return;
    }
    var folderpopup = document.getElementById('menu_SCCmenu_folder_context');
    if (!folderpopup) {
        log.error("Couldn't get 'menu_SCCmenu_folder'");
        return;
    }
    var em = ko.projects.extensionManager;
    em.addMenuItem(Components.interfaces.koIPart_SCC_file, document.createElement('menuseparator'));
    em.addMenuItem(Components.interfaces.koIPart_SCC_file, menupopup);
    em.addMenuItem(Components.interfaces.koIPart_SCC_collection, document.createElement('menuseparator'));
    em.addMenuItem(Components.interfaces.koIPart_SCC_collection, folderpopup);
    em.addMenuItem(Components.interfaces.koIPart_SCC_directory, document.createElement('menuseparator'));
    em.addMenuItem(Components.interfaces.koIPart_SCC_directory, folderpopup);
}

/* This takes an item, and determines what type of scc
   flags we are dealing with. returns true if we can do anything, or
   false if we cannot do anything */
peSCC.prototype.determineItemSCCSupport = function(/* koIPart */item, sccObj) {
    // prevent non-cvs directories from having to go through this torture
    var file = item.getFile();
    if (file.sccExclude) {
        return false;
    }
    // assume the worst =)
    sccObj.isscc_file = false;
    sccObj.isscc_directory = false;
    sccObj.isscc_container = false;
    sccObj.isscc_collection = false;

    var ok;
    try {
        ok = item.QueryInterface(Components.interfaces.koIPart_SCC_file) != null;
        if (ok) {
            sccObj.isscc_file = true;
        }
    } catch(e) {}
    try {
        ok = item.QueryInterface(Components.interfaces.koIPart_SCC_collection) != null;
        if (ok) {
            sccObj.isscc_collection = true;
        }
    } catch(e) {}
    try {
        ok = item.QueryInterface(Components.interfaces.koIPart_SCC_directory) != null;
        if (ok) {
            sccObj.isscc_directory = true;
        }
    } catch(e) {}
    sccObj.isscc_container = sccObj.isscc_directory || sccObj.isscc_collection;
    sccObj.isscc_file = sccObj.isscc_file || !sccObj.isscc_container;
    sccObj.has_scc = file.sccType != '';
    sccObj.is_scc = ('isscc_file' in sccObj && sccObj.isscc_file) || sccObj.isscc_container;
    if (!sccObj.is_scc) {
        return false;
    }

    sccObj.can_scc = sccObj.has_scc || file.sccDirType != '';
    sccObj.missing = !file.exists;

    var in_repository = null;
    if (item.url && !sccObj.can_scc && !sccObj.missing) {
        sccObj.can_scc = sccObj.has_scc || file.sccDirType;
    }

    if (file.sccExclude) return false;
    return true;
}

peSCC.prototype.supportsCommand = function(command, item) {
    // do we have either cvs or scc?
    if (!_prefs.getBooleanPref('cvsEnabled') &&
        !_prefs.getBooleanPref('p4Enabled') &&
        !_prefs.getBooleanPref('svnEnabled')) {
        return false;
    }
    // If scintilla has focus, should be talking about the file in scintilla
    if (!item) {
        if (ko.projects.getFocusedProjectView()) {
            item = ko.projects.active.getSelectedItem();
            // item must be a file
            if (!item || !item.url) {
                return false;
            }
        } else {
            var view = ko.views.manager.currentView;
            if (!view || !view.document || view.document.isUntitled) {
                // This is an unsaved file -- no SCC commands can apply.
                return false;
            }
            item = view.item;
            item.url = view.document.file.URI;
        }
    }
    // XXX look into cacheing sccObj into the item
    var sccObj = new Object();
    if (!this.determineItemSCCSupport(item,sccObj)) {
        return false;
    }
    var isfoldercmd = command.indexOf('_folder')>1;
    /* if cvs, and this is a directory, then the directory MUST be in cvs
    if (isfoldercmd) {
      dump("URL: "+item.url+"\n");
      ko.logging.dumpObject(sccObj);
      dump("  isfoldercmd "+isfoldercmd+"\n");
      dump("  folders_in_repository "+folders_in_repository+"\n");
    }
    */
    if (isfoldercmd && !sccObj.isscc_container) {
        return false;
    }
    /* folder handling is different between p4 and cvs */
    var file = item.getFile();
    var folders_in_repository = file.sccType == 'cvs' || file.sccDirType == 'cvs' ||
                        file.sccType == 'svn' || file.sccDirType == 'svn';
    var folders_revertable = file.sccType == 'p4' || file.sccDirType == 'p4';

    var foldercmd_ok = (sccObj.isscc_container && isfoldercmd) &&
            ((folders_in_repository && (sccObj.isscc_directory || sccObj.isscc_collection)) ||
            (!folders_in_repository && sccObj.can_scc));
    var retval;
    switch (command) {
    case 'cmd_SCCedit':
        retval = sccObj.has_scc &&
            (file.sccAction == '' || file.sccAction == 'integrate')
            && (file.sccType == 'p4' ||
                file.sccType == 'cvs' && file.isReadOnly)
        //dump('asked about '+ command + ', returning' + retval +'\n');
        return retval;
        break;
    case 'cmd_SCCadd_folder':
        retval = sccObj.can_scc && !sccObj.has_scc && isfoldercmd
                && sccObj.isscc_directory && folders_in_repository;
        //dump('asked about '+ command + ', returning' + retval +'\n');
        return retval;
        break;
    case 'cmd_SCCadd':
        retval = sccObj.can_scc && !sccObj.has_scc;
        //dump('asked about '+ command + ', returning' + retval +'\n');
        return retval;
        break;
    case 'cmd_SCCupdate_folder':
        return foldercmd_ok;
        break;
    case 'cmd_SCCupdate':
        retval = sccObj.has_scc && file.sccNeedSync;
        //dump('asked about '+ command + ', returning' + retval +'\n');
        return retval;
        break;
    case 'cmd_SCCremove':
        return sccObj.has_scc && !file.sccNeedSync &&
            file.sccAction == '';
        break;
    case 'cmd_SCCdiff_folder':
        return foldercmd_ok;
        break;
    case 'cmd_SCChistory':
        retval = sccObj.has_scc;
        return retval;
        break;
    case 'cmd_SCCdiff':
        retval = sccObj.has_scc &&(file.sccNeedSync ||
                file.sccAction == 'edit' || file.sccAction == 'integrate');
        //dump('asked about '+ command + ', returning' + retval +'\n');
        return retval;
        break;
    case 'cmd_SCCrevert_folder':
        return foldercmd_ok && folders_revertable;
        break;
    case 'cmd_SCCrevert':
        retval = sccObj.has_scc && (file.sccNeedSync ||
                file.sccAction != '');
        //dump('asked about '+ command + ', returning' + retval +'\n');
        return retval;
        break;
    case 'cmd_SCCcommit_folder':
        return foldercmd_ok;
        break;
    case 'cmd_SCCcommit':
        return sccObj.has_scc && file.sccAction != ''
            && (!file.sccNeedSync ||
                file.sccAction == 'delete');
        break;
    default:
        break;
    }
    return false;
}

peSCC.prototype.isCommandEnabled = peSCC.prototype.supportsCommand;

/* recursivly build a list of items that are being used in
   this scc operation */
peSCC.prototype._addVirtualFolderChildren = function(item)
{
    var items = [];
    try {
        var children = new Array();
        item.getChildren(children, new Object());
        children = children.value;
    } catch(e) {
        return items;
    }
    if (children.length < 1) return items;

    var child;
    for (var i = 0; i < children.length; i++) {
        child = children[i];
        if (child.url) {
            items.push(child);
        } else {
            var c = this._addVirtualFolderChildren(child);
            if (c.length > 0) {
                items = items.concat(c);
            }
        }
    }
    return items;
}

peSCC.prototype.doCommand = function(command) {
    window.setTimeout(function (self, command) { self._doCommand(command); }, 1, this, command);
}

peSCC.prototype._doCommand = function(command) {
    var cmd = null;
    var error = null;
    var urls=[];
    var _sccUrls = [];
    _sccUrls['p4'] = [];
    _sccUrls['cvs'] = [];
    _sccUrls['svn'] = [];
    var _sccLocalVersions = []; /* Only used for scc history */
    _sccLocalVersions['p4'] = [];
    _sccLocalVersions['cvs'] = [];
    _sccLocalVersions['svn'] = [];
    var _allURLS = [];
    var item = null;
    var items = [];
    var i;
    log.debug("doCommand "+command);
    // XXX FIXME this is an issue since projects do not necessarily have focus
    // on a context menu operation
    if (ko.projects.getFocusedProjectView()) {
        items = ko.projects.active.getSelectedItems();
    } else {
        //dump("Scintilla has focus -- actions should be based on the current file\n");
        var view = ko.views.manager.currentView;
        if (view && !view.document.isUntitled) {
            item = ko.projects.manager.findItemByURL(view.document.file.URI);
            if (!item) {
                item = view.item;
                item.url = view.document.file.URI;
                item.project = null;
            }
            items.push(item);
        }
    }

    if (items.length < 1) return;

    var isfoldercmd = command.indexOf('_folder')>1;
    for (i=0; i<items.length; i++) {
        // XXX look into cacheing sccObj into the item
        var sccObj = new Object();
        if (!this.determineItemSCCSupport(items[i],sccObj)) {
            log.debug("doCommand unable to determineItemSCCSupport "+items[i].url);
            return;
        }
        var file = items[i].getFile();
        var type = file.sccType;
        if (!type) type = file.sccDirType;
        if (!type) {
            log.debug("doCommand no sccType "+items[i].url);
            continue;
        }
        if (isfoldercmd && sccObj.isscc_collection && items[i].type == 'project') {
            var url = ko.uriparse.dirName(items[i].url)
            _sccUrls[type].push(url);
            _sccLocalVersions[type].push(""); /* No version, it's a collection */
            _allURLS.push(url);
        } else if (isfoldercmd && sccObj.isscc_collection) {
            if (typeof(items[i].url) != 'undefined' && items[i].url) {
                _sccUrls[type].push(items[i].url);
                _sccLocalVersions[type].push(""); /* No version, it's a collection */
                _allURLS.push(items[i].url);
            }
            // add the dialogs children
            items = items.concat(this._addVirtualFolderChildren(items[i]));
        } else {
            _sccUrls[type].push(items[i].url);
            _sccLocalVersions[type].push(file.sccLocalRevision);
            _allURLS.push(items[i].url);
        }

        // XXX implement virtual folder support
        // check for virtual folders, add the children of the virtual
        // folders to our items array.
        //if (sccObj.isscc_collection && items.children.length > 0) {
        //    var c = this._addVirtualFolderChildren(items[i]);
        //    if (c.length > 0) {
        //        items = items.concat(c);
        //    }
        //}
    }
    
    if ((command != 'cmd_SCCrevert') &&
        (command != 'cmd_SCChistory') &&
	!ko.views.manager.offerToSave(_allURLS, /* urls */
                      'Save Modified Files?', /* title */
                      'Save Modified Files Before Source Code Control Action', /* prompt */
                      "save_on_scc", /* pref */
                      false /* skip projects */
                      )) {
        return;
    }
    
    
    var output = '';
    var title;
    log.debug('selected for scc operation '+_sccUrls +'\n');
    for (var scchandler in sccModules) {
        if (!sccModules[scchandler]) continue; // skip unregistered SCC services
        if (_sccUrls[scchandler].length < 1) continue;
        urls = _sccUrls[scchandler];
        try {
            switch (command) {
            case 'cmd_SCCedit':
                window.setCursor("wait");
                output = sccModules[scchandler].edit(urls.length, urls);
                output = this.generateCommandInfo(scchandler+' edit',urls,output);
                title = this.summarizeSCCAction(scchandler, 'edit', urls);
                ko.scc.logger.reportResults(title, output);
                break;
            case 'cmd_SCCadd_folder':
                if (!isfoldercmd) return;
            case 'cmd_SCCadd':
                this.doAdd(urls, sccModules, scchandler);
                break;
            case 'cmd_SCCremove':
                this.doRemove(urls, sccModules, scchandler);
                break;
            case 'cmd_SCCrevert_folder':
                if (!isfoldercmd) return;
            case 'cmd_SCCrevert':
                this.doRevert(urls, sccModules, scchandler);
                break;
            case 'cmd_SCCupdate_folder':
                if (!isfoldercmd) return;
            case 'cmd_SCCupdate':
                this.doUpdate(urls, sccModules, scchandler);
                break;
            case 'cmd_SCCdiff_folder':
                if (!isfoldercmd) return;
            case 'cmd_SCCdiff':
                var options = '';
                var external = '';
                var diff;
                switch (scchandler) {
                case 'p4':
                    if (_prefs.getStringPref('p4_diff_type') == 'komododiff') {
                        options = _prefs.getStringPref('p4DiffOptions');
                    } else {
                        options = '';
                        external = _prefs.getStringPref('externaldiff');
                    }
                    break;
                case 'cvs':
                    options = _prefs.getStringPref('cvsDiffOptions');
                    break;
                case 'svn':
                    options = _prefs.getStringPref('svnDiffOptions');
                    break;
                default:
                    ko.dialogs.alert("Invalid Source Code Control Handler: "+scchandler);
                    return;
                }
                window.setCursor("wait");
                diff = sccModules[scchandler].diff(urls.length, urls, options, external);
                error = this._lastErrorSvc.getLastErrorMessage();
                //output = this.generateCommandInfo(scchandler+' diff',urls,diff);
                //ko.scc.logger.append(output);
                if (!error && external == '') {
                    if (diff) {
                        var diffDisplayStyle = _prefs.getStringPref('diffDisplayStyle');
                        if (diffDisplayStyle == 'tab') {
                            try {
                                var diffView = ko.views.manager.doNewView('Diff');
                                diffView.document.buffer = diff;
                            } catch (e) {
                                log.exception(e);
                            }
                        } else if (diffDisplayStyle == 'diffwin') {
                            // Put together a title for the diff output win.
                            title = scchandler + " diff:";
                            for (i=0; i < urls.length; i++) {
                                if (i > 0) title += ",";
                                title += " "+ko.uriparse.baseName(urls[i]);
                            }
                            ko.launch.diff(diff, title);
                        }
                    } else {
                        alert("Files are in sync with repository.");
                    }
                }
                break;
            case 'cmd_SCCcommit_folder':
                if (!isfoldercmd) return;
            case 'cmd_SCCcommit':
                this.doCommit(urls, sccModules, scchandler);
                break;
            case 'cmd_SCChistory':
                this.doHistory(urls, _sccLocalVersions[scchandler],
                               sccModules, scchandler);
                break;
            default:
                break;
            }
        } catch (e) {
            alert("An exception occurred during SCC operation. (" + e + ')');
            log.error(e);
        }
    }
    // Ensure the cached information gets updated, so the SCC toolbars and menu
    // are properly set.
    // Fixes bug: http://bugs.activestate.com/show_bug.cgi?id=48417
    ko.views.manager.resetLastViewCache();

    window.updateCommands('SCC'); // We've changed things hopefully, and no one else knows.
    window.setCursor("auto");
    error = this._lastErrorSvc.getLastErrorMessage();
    if (error)
        alert(error);
}

peSCC.prototype.generateCommandInfo = function(command,urls,text) {
    var out = command+'\n';
    for (var url in urls) {
        out += '    '+urls[url]+'\n'
    }
    out += text;
    return out;
}

peSCC.prototype.doAdd = function(urls, sccModules, scchandler) {
    window.setCursor("wait");
    var output = sccModules[scchandler].add(urls.length, urls,
                                        '', '');
    window.setCursor("auto");
    output = this.generateCommandInfo(scchandler+' add', urls, output);
    var title = this.summarizeSCCAction(scchandler, 'add', urls);
    ko.scc.logger.reportResults(title, output);
    return true;
}

peSCC.prototype.doRemove = function(urls, sccModules, scchandler) {
    var prompt = null;
    if (scchandler == 'p4') {
        prompt = "Select the file you would like to remove from the " +
                 "repository. (The selected files will be removed from " +
                 "your disk.)";
    } else {
        prompt = "Select the file you would like to remove from the " +
                 "repository.";
    }
    var selectedURLs = ko.dialogs.selectFromList(
            "Remove Files from Source Control", // title
            prompt,
            urls,
            null, // selectionCondition (null means default)
            ko.uriparse.displayPath); // stringifier
    if (selectedURLs == null) { // user cancelled dialog
        return false;
    }

    window.setCursor("wait");
    var force = true;
    var recursive = false;
    var output = sccModules[scchandler].remove(selectedURLs.length,
                                               selectedURLs, force,
                                               recursive);
    window.setCursor("auto");
    output = this.generateCommandInfo(scchandler+' remove', selectedURLs, output);
    var title = this.summarizeSCCAction(scchandler, 'remove', selectedURLs);
    ko.scc.logger.reportResults(title, output);
    return true;
}

peSCC.prototype.doRevert = function(urls, sccModules, scchandler) {
    log.debug("doRevert...");
  try {
    var selectedURLs = ko.dialogs.selectFromList(
            "Revert Files to Repository", // title
            "Select the files to revert to the repository version. "+
                "(Changes in opened documents will be lost).",
            urls,
            null, // selectionCondition (null means default)
            ko.uriparse.displayPath); // stringifier
    if (selectedURLs == null) { // user cancelled dialog
        return false;
    }

    window.setCursor("wait");
    try {
        var output = sccModules[scchandler].revert(selectedURLs.length, selectedURLs,'');
        output = this.generateCommandInfo(scchandler+' revert',selectedURLs,output);
        var title = this.summarizeSCCAction(scchandler, 'revert', selectedURLs);
        ko.scc.logger.reportResults(title, output);
    } catch (e) {
        log.error(e);
    }
    ko.views.manager.revertViewsByURL(selectedURLs);
    // handle reverting projects properly
    var isProjectFile;
    for (var i in selectedURLs) {
        isProjectFile = selectedURLs[i].slice(-4).toLowerCase() == '.kpf';
        if (isProjectFile) {
            ko.projects.manager.revertProjectByURL(selectedURLs[i]);
        }
    }
    window.setCursor("auto");
  } catch (e) {
    log.error(e);
  }
    return true;
}

peSCC.prototype.summarizeSCCAction = function(scchandler, action, urls)
{
    var retval = "Last command: " + action + ' ('+scchandler + ') on ';
    var display = ko.uriparse.displayPath(urls[0]);
    if (urls.length > 1) {
        retval = retval + ' ' + display + ', ... (' + urls.length + ' files/directories).';
    } else {
        retval = retval + ' ' + display + '.';
    }
    return retval;
}

peSCC.prototype.doCommit= function(urls, sccModules, scchandler) {
    var answer = sccdialog_commit(urls);
    if (answer == null) { // user cancelled
        return false;
    }

    window.setCursor("wait");
    var output = sccModules[scchandler].commit(answer.selected.length,
                                               answer.selected,
                                               answer.message, '');
    ko.views.manager.revertViewsByURL(answer.selected);
    window.setCursor("auto");
    output = this.generateCommandInfo(scchandler+' commit', answer.selected,
                                      output);
    var title = this.summarizeSCCAction(scchandler, 'commit', answer.selected);
    ko.scc.logger.reportResults(title, output);
    return true;
}

peSCC.prototype.doHistory = function(urls, localRevisions, sccModules, scchandler) {
    // Show the dialog.
    var obj = new Object();
    obj.urls = urls;
    obj.localRevisions = localRevisions;
    obj.sccType = scchandler;
    obj.sccHandler = sccModules[scchandler];
    window.openDialog(
        "chrome://komodo/content/scc/history.xul",
        "Komodo:SCCHistory",
        "chrome,titlebar,resizable=yes",
        obj);
    return true;
}

peSCC.prototype.doUpdate = function(urls, sccModules, scchandler) {
    window.setCursor("wait");
    var output = sccModules[scchandler].update(urls.length, urls,'');
    ko.views.manager.revertViewsByURL(urls);
    /* XXX force a refresh here, it will refresh selected url's.  This
       should be optimized later */
    ko.projects.refreshStatus();
    window.setCursor("auto");
    if (output) {
        this.handleChanges(output);
    }
    output = this.generateCommandInfo(scchandler+' update',urls,output);
    var title = this.summarizeSCCAction(scchandler, 'update', urls);
    ko.scc.logger.reportResults(title, output);
    return true;
}

peSCC.prototype.handleChanges = function(changes) {
    //XXX This pattern is specific to CVS update output. What about Perforce
    //    sync output? Is this a bug or is .handleChanges only called for CVS
    //    updates? --TrentM

    // re ^(?P<status>[UPARMC\?]) (?P<file>.*)$
    RegExp.multiline = true;
    var pattern = new RegExp("^([UCDP\?])\s+(.*?)\s+(.*?)$","g");
    var result;
    var conflicts = [];
    var haveConflict = false;
    var removed = [];
    var haveRemoved = false;
    var added = [];
    var haveNew = false;
    var file;
    var list;
    var i,j;
    while ((result = pattern.exec(changes)) != null) {
        //dump('found: '+result[1]+' '+result[2]+' '+result[3]+'\n');
        switch (result[1]) {
        case 'C':
            // conflict
            haveConflict = true;
            conflicts.push(result[3]);
            break;
        case 'D':
            // removed
            haveRemoved = true;
            removed.push(result[3]);
            break;
        case 'P':
            break; // Patched file can't be a new file, even though it may not be in our project.
        case 'U':
            // either updated or new file --
            // XXX need to check to make sure that it really is new -- we add
            // all files that aren't in the project
            if (!ko.projects.hasURL(result[3])) {
                haveNew = true;
                added.push(result[3]);
            }
            break;
        }
    }
    // look to see if files were added, ask if they should be
    // added to the current project.
    // XXX FIXME this needs to add for each project, not just the current
    var filepart, selectedURLs, obj;
    if (haveNew && ko.projects.manager.getCurrentProject()) {
        //XXX I think either (1) we should 'allowNone' here so the user could
        //    chose to not add new files, yet continue on to remove files, or
        //    better yet (2) we should group the added and removed files and
        //    ask the user to select which updates to carry to their project in
        //    one dialog.
        selectedURLs = ko.dialogs.selectFromList(
                "Add New Files", // title
                "Some new files from the repository are not in your "+
                    "project.  Select the files to add to your project.",
                added,
                null, // selectionCondition (null means default)
                ko.uriparse.displayPath); // stringifier
        if (selectedURLs == null) { // user cancelled dialog
            return false;
        }

        for (i = 0; i < selectedURLs.length; i++) {
            filepart = ko.projects.manager.getCurrentProject().createPartFromType('file');
            filepart.setStringAttribute('url', selectedURLs[i]);
            ko.projects.manager.addItem(filepart);
        }
    }
    // look to see if files were removed, notify of removal,
    // ask to remove from projects.
    if (haveRemoved) {
        var aroundToRemove = [];
        for (i = 0; i < removed.length; i++) {
            list = ko.projects.findPartsByURL(removed[i]);
            if (list.length > 0 && !list[0].live) {
                aroundToRemove.push(selectedURLs[i]);
            }
        }

        selectedURLs = ko.dialogs.selectFromList(
                "Remove Old Files", // title
                "Some files have been removed from the repository. "+
                    "Select the files to remove from your projects and "+
                    "toolbox.",
                aroundToRemove,
                null, // selectionCondition (null means default)
                ko.uriparse.displayPath); // stringifier
        if (selectedURLs == null) { // user cancelled dialog
            return false;
        }
        ko.projects.removeItemsByURLList(selectedURLs);
    }

    // look to see if there are conflicts, notify of conflicts,
    // and ask to open the files in komodo.  also mark any
    // items with conflict status
    if (haveConflict) {
        selectedURLs = ko.dialogs.selectFromList(
                "Merge Conflicts", // title
                "There are conflicts in some updated file(s). "+
                    "Would you like to open those file(s)?",
                conflicts,
                null, // selectionCondition (null means default)
                ko.uriparse.displayPath); // stringifier
        if (selectedURLs == null) { // user cancelled dialog
            return false;
        }
        ko.open.multipleURIs(selectedURLs);
    }

    // XXX reset the static var, is this necessary?
    RegExp.multiline = false;
    return true;
}

// this is hidden away now, no namespce, the registration keeps the reference
// we need
ko.projects.registerExtension(new peSCC());

this.editSCCFile = function editSCCFile(item)
{
    var urls = [];
    urls.push(item.url);
    window.setCursor("wait");
    var file = item.getFile();
    var lastErrorSvc = Components.classes["@activestate.com/koLastErrorService;1"]
                        .getService(Components.interfaces.koILastErrorService);
    lastErrorSvc.setLastError(0,"");
    try {
        sccModules[file.sccType].edit(urls.length, urls);
    } catch (e) {
        alert("An exception occurred during an SCC operation:" + e);
    }
    window.setCursor("auto");
    return lastErrorSvc.getLastErrorMessage();
}
}).apply(ko.projects);


ko.scc = {};
ko.scc.logger = {};
(function() {
/* SCC Scintilla control */

this.init = function SCCScinInit() {
    var scin = document.getElementById('scin-sccoutput');
    scin.initWithBuffer('', 'Text')
}

this.reportResults = function SCCScinReportResults(actiondescription, actiontext)
{
    document.getElementById('sccoutput-desc').setAttribute('value', actiondescription);
    ko.scc.logger.show();
    ko.scc.logger.append(actiontext);
}

this.show = function SCCTab_Show()
{
    var tabWidget = document.getElementById("sccoutput_tab");
    var tabsWidget = document.getElementById("output_tabs");
    if (! ko.uilayout.outputPaneShown() || tabsWidget.selectedItem != tabWidget) {
        if (ko.dialogs.yesNo("Show SCC Output Tab?", // prompt
                         "Yes", // response
                         null, // text
                         null, // title
                         'show_scc_tab') == "Yes") {
            ko.uilayout.ensureOutputPaneShown(window);
            tabsWidget.selectedItem = tabWidget;
        }
    }
}

this.append = function SCCScinAppendText(text) {
    if (typeof(text)=='undefined' || !text) return;
    var scin = document.getElementById('scin-sccoutput').scimoz;
    scin.readOnly = false;
    scin.currentPos = scin.textLength-1;
    // move cursor to end of current text
    text = "-------------------------------------------------------------------\n" +
        text+'\n';
    scin.addText(text.length, text);
    scin.gotoLine(scin.lineFromPosition(scin.textLength));
    scin.readOnly = true;
}

this.clear = function SCCScinClearText() {
    var scin = document.getElementById('scin-sccoutput').scimoz;
    scin.readOnly = false;
    scin.text = '';
    scin.readOnly = true;
}

}).apply(ko.scc.logger);

(function() {
this.warnSSHConfiguration = function SCCWarnAboutSSHIfNecessary() {
    var osSvc = Components.classes["@activestate.com/koOs;1"].getService();
    var CVS_RSH = osSvc.getenv('CVS_RSH');
    var SVN_SSH = osSvc.getenv('SVN_SSH');
    var msg, answer;
    
    // Only bother warning if CVS_RSH is set and CVS integration and CVS
    // background checking are enabled.
    var _prefs = Components.classes["@activestate.com/koPrefService;1"].
                    getService(Components.interfaces.koIPrefService).prefs;
    if (CVS_RSH
        && _prefs.getBooleanPref("cvsEnabled")
        && _prefs.getBooleanPref("cvsBackgroundCheck")
        // Checking this pref (seemingly redundantly) here is necessary
        // because the pref is being used to mean:
        //  "don't ask or do anything if true"
        // rather than the usual meaning:
        //  "don't ask, just presume user wants same action as last time"
        && !_prefs.getBooleanPref("donotask_cvs_ssh_setup_warning")
       )
    {
        msg = "The CVS_RSH environment variable is set to '"+CVS_RSH+
            "' and CVS Background Status Checking is currently turned "+
            "on.  Failure to properly configure CVS authentication may "+
            "result in process lockups within Komodo.  Please read the "+
            "Komodo documentation regarding the use of external "+
            "protocols, such as SSH, with Komodo.  \nIf you are sure "+
            "you have properly configured CVS/SSH authentication on "+
            "your machine, click 'Keep Enabled'.  Otherwise click "+
            "'Disable' to disable Komodo's CVS integration for now. "+
            "When you are sure you have setup CVS properly you can "+
            "re-enable CVS integration in Edit|Preferences|Source Control|CVS.";
        answer = ko.dialogs.customButtons(msg,
            ["Disable", "Disable and Tell Me More", "Keep Enabled"],
            "Disable", null, "Komodo CVS Integration",
            "cvs_ssh_setup_warning");
        switch (answer) {
            case "Keep Enabled":
                break;
            case "Disable and Tell Me More":
                ko.help.open("komodo-doc-scc.html#config_cvs");
                // fall through
            case "Disable":
            case "Cancel":
                _prefs.setBooleanPref("cvsEnabled", false);
        }
    }


    if (SVN_SSH
        && _prefs.getBooleanPref("svnEnabled")
        && _prefs.getBooleanPref("svnBackgroundCheck")
        // Checking this pref (seemingly redundantly) here is necessary
        // because the pref is being used to mean:
        //  "don't ask or do anything if true"
        // rather than the usual meaning:
        //  "don't ask, just presume user wants same action as last time"
        && !_prefs.getBooleanPref("donotask_svn_ssh_setup_warning")
       )
    {
        msg = "The SVN_SSH environment variable is set to '"+SVN_SSH+
                  "' and Subversion Background Status Checking is "+
                  "currently turned on.  Failure to properly "+
                  "configure Subversion authentication may result "+
                  "in process lockups within Komodo.  Please read "+
                  "the Komodo documentation regarding the use of "+
                  "external protocols, such as SSH, with Komodo.  \n"+
                  "If you are sure you have properly configured "+
                  "Subversion/SSH authentication on your machine, "+
                  "click 'Keep Enabled'.  Otherwise click 'Disable' "+
                  "to disable Komodo's Subversion integration for "+
                  "now. When you are sure you have setup Subversion "+
                  "properly you can re-enable Subversion integration "+
                  "in Edit|Preferences|Source Control|Subversion.";
        answer = ko.dialogs.customButtons(msg,
            ["Disable", "Disable and Tell Me More", "Keep Enabled"],
            "Disable", null, "Komodo Subversion Integration",
            "svn_ssh_setup_warning");
        switch (answer) {
            case "Keep Enabled":
                break;
            case "Disable and Tell Me More":
                ko.help.open("komodo-doc-scc.html#config_svn");
                // fall through
            case "Disable":
            case "Cancel":
                _prefs.setBooleanPref("svnEnabled", false);
        }
    }
}

}).apply(ko.scc);
